# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import weakref
import logging
from typing import Optional
from . import Killable


class Process(Killable):
    """An object encapsulating a process within a container.
    Do not instantiate directly, use container.spawn_process."""
    def __init__(self, parent, uuid, data_callback, termination_callback, stderr_callback=None):
        super().__init__()
        self.parent = weakref.ref(parent)
        self.node = weakref.ref(parent.parent())
        self.location = weakref.ref(self.node().parent())
        self.conn = weakref.ref(parent.conn())
        self.uuid = uuid
        self.data_callback = data_callback
        self.termination_callback = termination_callback
        self.stderr_callback = stderr_callback
        self.wrapper = None

    def stdin(self, data: bytes):
        """Inject data into stdin for the process.

        :param data: The data to inject -  bytes, not a string.

        Note that because this injects raw data, it may not behave as you expect. Remember to:

        * Turn strings into bytes with .encode()
        * Add '\\\\n' to emulate carriage return.
        * Turn returned bytes into strings with .decode()

        """
        # a client side disconnection?
        self.ensure_alive()
        if len(data) == 0:
            self.destroy()
            return

        self.conn().send_cmd(b'stdin_process', {'node': self.node().pk,
                                                'container': self.parent().uuid,
                                                'process': self.uuid}, bulk=data)

    def destroy(self, with_command=True):
        if self.bail_if_dead():
            return

        # and then wait for notice of termination in msg.bulk
        if with_command:
            self.conn().send_cmd(b'destroy_process', {'node': self.node().pk,
                                                      'container': self.parent().uuid,
                                                      'process': self.uuid})
        logging.info("Terminated client side: %s" % self.uuid.decode())
        self.mark_as_dead()
        if self.termination_callback is not None:
            self.termination_callback(self, 0)

    def give_me_messages(self, msg):
        if self.bail_if_dead():
            return

        if msg is None:
            logging.error("give_me_messages was sent None")
            return

        # Has the process died?
        if len(msg.bulk) == 0:
            logging.info("Terminated server side: %s (%d)" % (self.uuid.decode(), msg.params['returncode']))
            self.mark_as_dead()
            if self.termination_callback is not None:
                self.location().call_on_main(self.termination_callback, (self, msg.params['returncode'],))
            return

        # Stderr?
        if 'stderr' in msg.params:
            if self.stderr_callback is not None:
                self.stderr_callback(self, msg.bulk)
            return

        # Otherwise we're just data
        if self.data_callback is not None:
            self.data_callback(self, msg.bulk)

    def __repr__(self):
        return "<Process '%s'>" % self.uuid.decode()



# Copyright (c) 2018 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import logging
import shortuuid
from secrets import token_urlsafe
from tfnz.location import Location
from tfnz.volume import Volume
from tfnz.components.postgresql import Postgresql
from tfnz.endpoint import WebEndpoint, Cluster


class Mezzanine:
    """Puts a Python/Django/Mezzanine instance on each node and load balances. Hosts wsgi applications.

    :param location: A location (object) to connect to.
    :param volume: A volume (object) to use as a persistent store - will be mounted on static/media/uploads
    :param sql_volume: A volume to connect to a Postgres server for SQL storage.
    :param db: An instance of a Postgresql component.
    :param fqdn: The FQDN to publish to.
    :param app_name: Name of the application (used to create file paths).
    :param image: The image to use (assumed derived 'FROM tfnz/mezzanine')
    :param log_callback: An optional callback for log messages -  signature (object, bytes)
    :param superuser: if creating from fresh set this (username, email) as the admin user
    :param debug: Whether or not to set debug on django."""

    def __init__(self, location: Location, volume: Volume, sql_volume: Volume, fqdn: str,
                 app_name: str, image, *, log_callback=None, superuser=None, debug=False):
        nodes = location.ranked_nodes()
        manage = 'python3 /%s/manage.py ' % app_name

        # spawn database (initialises async)
        self.db = Postgresql(nodes[0], sql_volume)

        # create the first webserver and use that to initialise a bunch of things
        first_server = nodes[0].spawn_container('tfnz/mezzanine' if image is None else image,
                                                volumes=[(volume, '/%s/static/media/uploads/' % app_name)])
        localsettings = Mezzanine.localsettings_template % \
                        (self.db.password, self.db.private_ip(), str(debug), token_urlsafe(32), token_urlsafe(32))
        first_server.put('/%s/%s/local_settings.py' % (app_name, app_name), localsettings.encode())
        self.db.allow_connection_from(first_server)

        # do we need to initialise?
        # ensure database is also wait_until_ready
        if self.db.ensure_database('mezzanine'):
            first_server.run_process(manage + 'createdb --noinput --nodata', data_callback=log_callback)
            if superuser is not None:
                first_server.run_process(manage + 'createsuperuser --noinput --username %s --email %s'
                                         % (superuser[0], superuser[1]), data_callback=log_callback)

        # start additional webservers
        self.webservers = [first_server]
        for node in nodes[1:]:
            server = node.spawn_container('tfnz/mezzanine' if image is None else image,
                                          volumes=[(volume, '/%s/static/media/uploads/' % app_name)])
            server.put('/%s/%s/local_settings.py' % (app_name, app_name), localsettings.encode())
            self.webservers.append(server)

        # configure and go
        for w in self.webservers:
            self.db.allow_connection_from(w)
            w.run_process('rm /etc/nginx/conf.d/default.conf')
            w.run_process('mkdir /run/nginx')
            w.put('/etc/nginx/conf.d/nginx.conf', (Mezzanine.nginx_template % (fqdn, '/%s/' % app_name)).encode())
            w.run_process(manage + 'collectstatic --noinput')
            w.spawn_process('cd %s ; gunicorn -b unix:/tmp/gunicorn.sock --workers=8 %s.wsgi'
                            % (app_name, app_name), stderr_callback=log_callback)
            w.run_process('nginx')

        # gather together and serve into an endpoint
        self.cluster = Cluster(self.webservers)
        location.endpoint_for(fqdn).publish(self.cluster, fqdn)

        # wait until we're actually able to serve
        WebEndpoint.wait_http_200(fqdn)
        logging.info("Mezzanine is up.")

    def change_password(self, username, password):
        """change the username and password of the given user - could be superuser"""
        # also, this is a bit gross
        w = self.webservers[0]
        py = Mezzanine.chpass_template % (password, shortuuid.uuid(), username, self.db.private_ip())
        w.put('update_password.py', py.encode())
        w.run_process('python3 update_password.py')
        w.run_process('rm update_password.py')

    localsettings_template = """
DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.postgresql_psycopg2",
        "NAME": "mezzanine",
        "USER": "postgres",
        "PASSWORD": "%s",
        "HOST": "%s",
        "PORT": "5432"
    }
}

ALLOWED_HOSTS = ["*"]

DEBUG = %s

SECRET_KEY = "%s"
NEVERCACHE_KEY = "%s"
"""

    nginx_template = """
upstream gunicorn {
    server unix:/tmp/gunicorn.sock fail_timeout=0;
}    

server {
    listen 80;
    server_name %s;
    charset utf-8;

    location /static/ {
        root %s;
    }

    location / {
        proxy_pass http://gunicorn;
    }
}
"""
    chpass_template = """
from django.contrib.auth.hashers import PBKDF2PasswordHasher
import psycopg2

hash = PBKDF2PasswordHasher().encode('%s', '%s')
sql = "UPDATE auth_user SET password='" + hash + "' WHERE username='%s'"

conn = psycopg2.connect("dbname=mezzanine user=postgres host=%s")
cur = conn.cursor()
cur.execute(sql)
conn.commit()
cur.close()
conn.close()
"""



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.


# don't rename this as tfnz since it will clash with the tfnz module

import sys
import re
import os
import os.path
import argparse
from sys import argv, stderr
from messidge import default_location, create_account
from tfnz.location import Location
from tfnz.volume import Volume
from tfnz.docker import Docker
from tfnz.endpoint import Cluster
from tfnz.cli import base_argparse, systemd, Interactive


def main_impl():
    # see if we have an account
    try:
        default_location(prefix="~/.20ft")
    except RuntimeError:
        # there is no default location
        if len(argv) != 3:
            print("""
Either there is no a 20ft account on this machine, and you need 
to request access from your administrator; or the ~/.20ft/default_location
file has been deleted. Run something like...

    echo 'location.20ft.nz' > ~/.20ft/default_location

If you already have an account on another machine you can merely 
copy the directory ~/.20ft (and it's contents) to this machine.""", file=stderr)
            return None

        # trying to create a new account using token passed by the server
        create_account(argv[1], argv[2], prefix="~/.20ft")
        print("Created OK for: %s\n" % argv[1])
        return None

    # right, get on with it
    parser = base_argparse('tfnz')
    launch_group = parser.add_argument_group('launch options')
    launch_group.add_argument('-v', '--verbose', help='verbose logging', action='store_true')
    launch_group.add_argument('-q', '--quiet', help='no logging', action='store_true')
    launch_group.add_argument('-ti', '--interactive', help='interactive, connect stdin and stdout', action='store_true')
    launch_group.add_argument('-e', '--env', help='set an environment variable, possibly from current',
                              action='append', metavar='[VAR=value | VAR]')
    launch_group.add_argument('-f', '--file', help='write a pre-boot file', action='append', metavar='src:dest')
    launch_group.add_argument('-m', '--mount', help='mount a volume', action='append',
                              metavar='[uuid:mountpoint | tag:mountpoint]')
    launch_group.add_argument('-p', '--publish', help='add a local->remote tcp proxy', action='append',
                              metavar='8080:80')
    launch_group.add_argument('-w', '--web', help='publish on web endpoint',
                              metavar='subdomain.my.com[:www.my.com[:certname]]')
    development_group = parser.add_argument_group('development options')
    development_group.add_argument('--ssh', help='create an ssh/sftp wrapped shell on given port',
                                   metavar='2222')
    development_group.add_argument('-s', help='shorthand for --ssh 2222', action='store_true')
    development_group.add_argument('-z', '--sleep', help='launch the container asleep (instead of entrypoint)',
                                   action='store_true')
    server_group = parser.add_argument_group('server options')
    server_group.add_argument('--systemd', help='create a systemd service', metavar='user@server.my.com')
    server_group.add_argument('--identity', help='specify an identity file to use with --systemd',
                              metavar="~/.ssh/some_id.pem")

    parser.add_argument('source', help="if '.' runs the most recently added docker image; "
                                       "else this is the tag or hex id of an image to run.")

    parser.add_argument('command', help='run this command/entrypoint instead', nargs='?')
    parser.add_argument('args', help='arguments to pass to a script or subprocess', nargs=argparse.REMAINDER)

    args = parser.parse_args()

    # collect any pre-boot files
    preboot = []
    if args.file is not None:
        for e in args.file:
            match = re.match('^[\w:]*[\w.\-\\/]+:[\w.\-\\/]+$', e)
            if not match:
                print("Pre-boot copies need to be in source:destination pairs", file=sys.stderr)
                print("....error in '%s'" % e, file=sys.stderr)
                return None
            files = e.split(':')
            # are we copying into a directory? append the original filename
            if files[1][-1:] == '/':
                files[1] += files[0]
            # try to read the file in
            try:
                with open(files[0], 'rb') as f:
                    preboot.append((files[1], f.read()))
            except FileNotFoundError:
                print("Could not find the source pre-boot file: " + files[0], file=sys.stderr)
                return None

    # if publishing to an endpoint, we might need ssl certs
    endpoint = None
    rewrite = None
    cert = None
    fqdns = None
    if args.web is not None:
        # split into endpoint:rewrite:certname (rewrite and certname are optional)
        fqdns = args.web.split(':')
        if len(fqdns) == 0 or len(fqdns[0]) == 0:
            print("Cannot publish to an endpoint without an address", file=sys.stderr)
            return None
        endpoint = fqdns[0]

        if len(fqdns) > 1 and len(fqdns[1]) > 0:
            rewrite = fqdns[1]

        if len(fqdns) > 2 and len(fqdns[2]) > 0:
            cert = (fqdns[2] + '.crt', fqdns[2] + '.key')
            if not os.path.exists(cert[0]):
                print("Cannot find certificate for ssl: " + cert[0], file=sys.stderr)
                return None
            if not os.path.exists(cert[1]):
                print("Cannot find key for ssl: " + cert[1], file=sys.stderr)
                return None

    # connect
    location = None
    try:
        location = Location(args.location, location_ip=args.local, quiet=args.quiet, debug_log=args.verbose)
    except BaseException as e:
        print("Failed while connecting to location: " + str(e), file=sys.stderr)
        return location

    # are we making a systemd service?
    if args.systemd is not None:
        systemd(location, args, argv, preboot, cert)
        return location

    # are we going to be using the most recent build?
    if args.source == '.':
        args.source = Docker.last_image()

    # create env-vars
    e_vars = set()
    environment = []
    if args.env is not None:
        for e in args.env:
            match_single = re.match('^[[0-9A-Za-z:_]+$', e)
            if match_single:
                if e not in os.environ:
                    print("Environment variable \'%s\' was not found in the local environment.", file=sys.stderr)
                    return location
                e += '=\'%s\'' % os.environ[e]
            match_extended = re.match('^[0-9A-Za-z:_]+=', e)
            if not match_extended:
                print("Environment variables need to be passed as 'name' or 'name=value' pairs", file=sys.stderr)
                print("....error in '%s'" % e, file=sys.stderr)
                return location
            variable = match_extended.group(0)[:-1]
            if variable in e_vars:
                print("Can only pass one value per environment variable.", file=sys.stderr)
                print("....error in '%s'" % e, file=sys.stderr)
                return location
            value = e[len(variable)+1:]
            e_vars.add(variable)
            environment.append((variable, value))

    # create portmaps
    l_ports = set()
    portmap = []
    if args.publish is not None:
        for e in args.publish:
            match = re.match('\d+:\d+$', e)
            if not match:
                print("Portmaps need to be passed as number:number pairs", file=sys.stderr)
                print("....error in '%s'" % e, file=sys.stderr)
                return location
            local, remote = e.split(':')
            if local in l_ports:
                print("Cannot bind a local port twice.", file=sys.stderr)
                print("....error in '%s'" % e, file=sys.stderr)
                return location
            l_ports.add(local)
            portmap.append((local, remote))

    # create volume mounts ensuring they don't overlap
    volumes = []
    mount_points = set()
    if args.mount is not None:
        for m in args.mount:
            if ':' not in m:
                print("Volumes need to be passed as uuid:mountpoint pairs", file=sys.stderr)
                print("....error in '%s'" % m, file=sys.stderr)
                return location
            find = m.rfind(':')
            key = m[:find]
            mount = m[find+1:]
            intersection = Volume.trees_intersect(mount_points, mount)
            if intersection is not None:
                print("Error in volumes: %s is a subtree of %s" % (intersection[0], intersection[1]), file=sys.stderr)
                return location
            try:
                volumes.append((location.volumes.get(location.user_pk, key), mount))
            except KeyError:
                print("Could not find volume: " + key)
                return location
            mount_points.add(mount)

    # overloaded command? args? (makes it into a string)
    if args.command is not None and args.args is not None:
        args.command = [args.command]
        args.command.extend(args.args)

    # try to launch the container
    try:
        node = location.node()
        container = node.spawn_container(args.source,
                                         env=environment,
                                         pre_boot_files=preboot,
                                         volumes=volumes,
                                         stdout_callback=(Interactive.stdout_callback if args.interactive
                                                          else lambda _, out: print(out.decode(), flush=True)),
                                         termination_callback=location.stop,
                                         command=args.command,
                                         sleep=args.sleep)
        container.wait_until_ready()  # a transport for exceptions
    except BaseException as e:
        print("Failed while spawning container: " + str(e), file=sys.stderr)
        return location

    # create the tunnels
    try:
        for m in portmap:
            container.attach_tunnel(m[1], m[0])
    except OSError as e:
        print("Failed while creating a tunnel onto the container: " + str(e))
        return location

    # launch an ssh server? user/pass are anything/anything - you can only connect from localhost anyway
    try:
        if args.ssh or args.s:
            container.create_ssh_server(2222 if args.s else int(args.ssh))
    except OSError as e:
        print("Failed while creating an ssh server onto the container: " + str(e))
        return location

    # publish to an endpoint?
    try:
        if args.web is not None:
            container.wait_until_ready()
            clstr = Cluster([container], rewrite=rewrite)
            ep = location.endpoint_for(endpoint)
            ep.publish(clstr, fqdns[0], ssl=cert)
    except ValueError as e:
        print("Failed while publishing to an endpoint: " + str(e))
        return location

    # running interactively?
    interactive = None
    if args.interactive:
        interactive = Interactive(location, container)

    # go
    location.run()
    if interactive is not None:
        interactive.stop()

    return None


def main():
    maybe_location = main_impl()
    if maybe_location is not None:
        maybe_location.disconnect()


if __name__ == "__main__":
    main()



#!/usr/local/bin/python3.5
# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import sys
from tfnz.cli import generic_cli, base_argparse


def main():
    parser = base_argparse('tfvolumes')
    subparsers = parser.add_subparsers(title='commands', dest='command')
    p_list = subparsers.add_parser('list', help='list available volumes')
    p_create = subparsers.add_parser('create', help='create a volume')
    p_create.add_argument('--sync', action='store_true', help='force synchronous updates')
    p_create.add_argument('tag', nargs='?', help='give the volume a tag')
    p_delete = subparsers.add_parser('destroy', help='destroy a volume')
    p_delete.add_argument('uuid')

    generic_cli(parser, {'list': list_vol, 'create': create_vol, 'destroy': destroy_vol})


def list_vol(location, args):
    for vol in location.all_volumes():
        print(vol.namespaced_display_name())


def create_vol(location, args):
    vol = location.create_volume(tag=args.tag, async=not args.sync)
    print(vol.namespaced_display_name())


def destroy_vol(location, args):
    try:
        vol = location.volumes.get(location.user_pk, key=args.uuid)
        location.destroy_volume(vol)
    except KeyError:
        print("Can't find volume: " + args.uuid)
        exit(1)


if __name__ == "__main__":
    main()



#!/usr/local/bin/python3.5
# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""A tiny utility to select from the location keys stored in ~/.20ft/"""

import os.path
from messidge import default_location
from tfnz.location import Location
from tfnz.cli import generic_cli, base_argparse


def main():
    parser = base_argparse('tflocations', location=False)
    subparsers = parser.add_subparsers(title='commands', dest='command')
    p_list = subparsers.add_parser('list', help='list locations with an account')
    p_token = subparsers.add_parser('select', help='select which location to use as default_location')
    p_token.add_argument('location', metavar='location.20ft.nz')

    generic_cli(parser, {'list': list_locations, 'select': select_location}, location=False)


def list_locations(location, args):
    # we have a match
    dl = default_location("~/.20ft")
    for loc in Location.all_locations():
        print("%s %s" % (loc, '<== default' if dl == loc else ''))


def select_location(location, args):
    # do a 'substring from left' match
    length = len(args.location)
    for loc in Location.all_locations():
        if loc[:length] == args.location:
            # we have a match
            with open(os.path.expanduser('~/.20ft/default_location'), 'w') as f:
                f.write(loc + "\n")
                return
    # failed
    print("Could not find a location starting with: " + args.location)
    exit(1)


if __name__ == "__main__":
    main()



#!/usr/local/bin/python3.5
# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import sys
from tfnz.cli import generic_cli, base_argparse


def main():
    parser = base_argparse('tfdomains')
    subparsers = parser.add_subparsers(title='commands', dest='command')
    p_list = subparsers.add_parser('list', help='list domains')
    p_token = subparsers.add_parser('prepare', help='receive a pre-claim token')
    p_token.add_argument('prepare_domain', metavar='my.com')
    p_create = subparsers.add_parser('claim', help='claim a domain')
    p_create.add_argument('claim_domain', metavar='my.com')
    p_global = subparsers.add_parser('global', help='make a domain global')
    p_global.add_argument('global_domain', metavar='my.com')
    p_private = subparsers.add_parser('private', help='make a domain private')
    p_private.add_argument('private_domain', metavar='my.com')
    p_release = subparsers.add_parser('release', help='release your claim')
    p_release.add_argument('release_domain', metavar='my.com')

    generic_cli(parser, {'list': list_dom, 'prepare': prepare, 'claim': claim,
                         'private': private, 'global': gbl, 'release': release})


def list_dom(location, args):
    for dom in location.endpoints.values():
        print(dom.domain)


def prepare(location, args):
    rtn = location.conn.send_blocking_cmd(b'prepare_domain', {'domain': args.prepare_domain})
    print("Put a DNS record on your domain: tf-token.%s, TXT=%s" %
          (args.prepare_domain, rtn.params['token'].decode()))
    print("...then run: tfdomains claim " + args.prepare_domain)
    print("The request will time out (and become invalid) after six hours.")


def claim(location, args):
    location.conn.send_blocking_cmd(b'claim_domain', {'domain': args.claim_domain})
    print("Claimed successfully - you can remove the tf-token record from DNS")


def gbl(location, args):
    location.conn.send_blocking_cmd(b'make_domain_global', {'domain': args.global_domain})
    print("Domain made global, clients will need to re-attach to see the change")


def private(location, args):
    location.conn.send_blocking_cmd(b'make_domain_private', {'domain': args.private_domain})
    print("Domain made private, clients will need to re-attach to see the change but can no longer publish.")


def release(location, args):
    location.conn.send_blocking_cmd(b'release_domain', {'domain': args.release_domain})
    print("Released domain")


if __name__ == "__main__":
    main()



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.


import logging
import weakref
import shortuuid
from base64 import b64encode
from typing import List, Optional, Tuple
from messidge.client.message import Message
from .docker import Docker
from .container import Container
from .volume import Volume


class Node:
    """An object representing a single node. Do not construct directly, use ranked_nodes on the location."""
    def __init__(self, parent, pk, conn, stats):
        # Internal Use: Don't construct nodes directly
        self.parent = weakref.ref(parent)
        self.pk = pk
        self.conn = weakref.ref(conn)
        self.stats = stats  #: A dictionary of performance stats for this node. Automatically kept up to date.
        self.containers = {}  #: A uuid->object map of containers running on this node (for the current session).

    def spawn_container(self, image: str, *,
                        env: Optional[List[Tuple[str, str]]]=None,
                        sleep: Optional[bool]=False,
                        volumes: Optional[List[Tuple[Volume, str]]]=None,
                        pre_boot_files: Optional[List[Tuple[str, bytes]]]=None,
                        command: Optional[str]=None,
                        stdout_callback: Optional[object]=None,
                        termination_callback: Optional[object]=None,
                        advertised_tag: Optional[str]=None) -> Container:
        """Asynchronously (by default) spawns a container on the node.

        :param image: the short image id from Docker.
        :param env: a list of environment name, value pairs to be passed.
        :param sleep: replaces the Entrypoint/Cmd with a single blocking command (container still boots).
        :param volumes: a list of (volume, mountpoint) pairs.
        :param pre_boot_files: List (filename, data) pairs to write into the container before booting.
        :param command: ignores Entrypoint/Cmd and launches with this script instead, list not string.
        :param stdout_callback: Called when the container sends output - signature (container, string).
        :param termination_callback: For when the container completes - signature (container, returncode).
        :param advertised_tag: A tag used so other sessions (for the same user) can reference the container.
        :return: A Container object.

        The resulting Container is initially a placeholder until the container has spawned.
        Any layers that need to be uploaded to the location are uploaded automatically.
        Note that the container will not be marked as ready until it actually has booted.

        To launch synchronously call wait_until_ready() on the container. This will also cause any exceptions
        to be transported to the calling thread."""

        # Ensure the lists are lists of tuples (or None)
        if env is not None and len(env) > 0:
            try:
                if len(env[0]) != 2:
                    raise TypeError  # do this because env[0] throws the same if we pass the wrong type
            except TypeError:
                raise ValueError("You need to pass a list of tuples for env vars - [(variable_name, data), ...]")
        if volumes is not None and len(volumes) > 0:
            try:
                for vol in volumes:
                    if len(vol) != 2:
                        raise TypeError
            except TypeError:
                raise ValueError("You need to pass a list of tuples for volumes - [(volume_object, mount_point), ...]")
        if pre_boot_files is not None and len(pre_boot_files) > 0:
            try:
                if len(pre_boot_files[0]) != 2:
                    raise TypeError
            except TypeError:
                raise ValueError("You need to pass a list of tuples for pre-boot files - [(filename, data), ...]")
        if advertised_tag is not None:
            # will throw an exception if it's no good
            reply = self.conn().send_blocking_cmd(b'approve_tag', {'user': self.parent().user_pk,
                                                                   'tag': advertised_tag})

        # Make it go...
        descr = Docker.description(image, self.conn())

        if command is not None:
            descr['Config']['Entrypoint'] = []
            descr['Config']['Cmd'] = command  # will overwrite the container's config
        vol_struct = [(vol[0].uuid, vol[1]) for vol in volumes] if volumes is not None else None
        layers = self.parent().ensure_image_uploaded(image, descr=descr)

        # Create the container object then tell the node to actually create it
        uuid = shortuuid.uuid().encode()
        self.containers[uuid] = Container(self, image, uuid, descr, env, volumes,
                                          stdout_callback=stdout_callback, termination_callback=termination_callback)
        logging.info("Spawning container: " + uuid.decode())
        cookie = {'session': self.conn().rid, 'user': self.parent().user_pk, 'tag': advertised_tag}
        self.conn().send_cmd(b'spawn_container', {'node': self.pk,
                                                  'layer_stack': layers,
                                                  'description': descr,
                                                  'env': env,
                                                  'volumes': vol_struct,
                                                  'pre_boot_files': pre_boot_files,
                                                  'sleep': sleep,
                                                  'cookie': cookie},
                             uuid=uuid,
                             reply_callback=self._container_status_update)
        return self.containers[uuid]

    def destroy_container(self, container: Container):
        """Destroy a container running on this node. Will also destroy any tunnels onto the container.

        :param container: The container to be destroyed."""
        container.ensure_alive()
        container._internal_destroy()
        # removing from .containers and marking any volumes as being free happens in container_status_update

    def all_containers(self) -> List[Container]:
        """Returns a list of all the containers running on this node (for *this* session)

        :return: A list of Container objects."""
        return list(self.containers.values())

    def _internal_destroy(self):
        """Called when the node needs to clean itself up"""
        # fake container status update messages
        for ctr in self.containers.values():
            msg = Message()
            msg.uuid = ctr.uuid
            msg.params = {'status': 'destroyed'}
            self._container_status_update(msg)

    def _update_stats(self, stats):
        # the node telling us it's current resource state
        self.stats = stats

    def _container_status_update(self, msg):
        try:
            container = self.containers[msg.uuid]
        except KeyError:
            logging.debug("Status update was sent for a non-existent container")
            return

        if 'exception' in msg.params:
            self.parent().raise_on_main(ValueError(msg.params['exception']))
            container.mark_as_ready()
            return

        if 'status' not in msg.params:
            if container.stdout_callback is not None:
                container.stdout_callback(container, msg.bulk)
            return

        if msg.params['status'] == 'running':
            logging.info("Container is running: " + msg.uuid.decode())
            logging.info("...startup time was: %.3fs" % msg.params['startup_time'])
            container.ip = msg.params['ip']
            container.mark_as_ready()
            return

        if msg.params['status'] == 'destroyed':
            # wait lock will still be locked if the container did not successfully start
            if container.wait_lock.locked():
                self.parent().raise_on_main(ValueError("Container did not manage to start"))
                container.mark_as_ready()  # to unblock the lock if nothing else

            # destroy
            container._internal_destroy(send_cmd=False)
            del self.containers[msg.uuid]

    def __repr__(self):
        return "<Node '%s' containers=%d>" % \
               (b64encode(self.pk).decode(), len(self.containers))



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

from setuptools import setup, find_packages

setup(name='tfnz',
      version='1.2.26',
      author='David Preece',
      author_email='davep@20ft.nz',
      url='https://20ft.nz',
      license='BSD',
      packages=find_packages(exclude=["messidge*", "docs*", "build*"]),
      install_requires=['pyzmq', 'libnacl', 'py3dns', 'requests', 'shortuuid', 'cbor',
                        'paramiko', 'psutil', 'requests_unixsocket', 'bottle', 'messidge'],
      description='SDK for 20ft.nz',
      long_description="The SDK for the 20ft.nz container PaaS. " +
                       "This package contains the SDK, command line tools and man pages. " +
                       "Main documentation is at http://docs.20ft.nz",
      keywords='container containers PaaS docker orchestration 20ft 20ft.nz',
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'Intended Audience :: Developers',
          'Intended Audience :: Information Technology',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: BSD License',
          'Natural Language :: English',
          'Topic :: Software Development :: Testing',
          'Topic :: System :: Software Distribution',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6'
      ],
      entry_points={
          'console_scripts': ['tfnz=tfnz.cli.tf:main',
                              'tfvolumes=tfnz.cli.tfvolumes:main',
                              'tfdomains=tfnz.cli.tfdomains:main',
                              'tfacctbak=tfnz.cli.tfacctbak:main',
                              'tflocations=tfnz.cli.tflocations:main',
                              'tfresources=tfnz.cli.tfresources:main',
                              'tfcache=tfnz.cli.tfcache:main',
                              'tfdescribe=tfnz.cli.tfdescribe:main']
      }
      )



#!/usr/local/bin/python3.5
# Copyright (c) 2017 David Preece, All rights reserved.
# 
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
# 
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import sys
import os.path
from messidge import default_location, KeyPair

bash_template = """echo 'mkdir -p ~/.20ft
cat > ~/.20ft/%s << EOF
%s
EOF
cat > ~/.20ft/%s.pub << EOF
%s
EOF
cat > ~/.20ft/%s.spub << EOF
%s
EOF
cat > ~/.20ft/default_location << EOF
%s
EOF

chmod 400 ~/.20ft/%s*' | /bin/sh
"""


def main():
    loc = default_location(prefix="~/.20ft") if len(sys.argv) == 1 else sys.argv[1]
    kp = KeyPair(loc, prefix="~/.20ft")
    with open(os.path.expanduser('~/.20ft/%s.spub' % loc)) as f:
        spub = f.read()
    print(bash_template % (loc, kp.secret.decode(), loc, kp.public.decode(), loc, spub, loc, loc))


if __name__ == "__main__":
    main()



#!/usr/local/bin/python3.5
# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import json
from tfnz.cli import generic_cli, base_argparse
from tfnz.docker import Docker


def main():
    parser = base_argparse('tfdescribe', location=False)
    parser.add_argument('image', help='image UUID or tag to describe')
    generic_cli(parser, {None: describe_image}, quiet=False, location=False)


def describe_image(location, args):
    desc = Docker.description(args.image)
    print(json.dumps(desc, indent=2))


if __name__ == "__main__":
    main()



# Copyright (c) 2018 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import sys
import argparse
import termios  # this *is* used and *can* be found
import tty
import select
import re
import logging
import os
from threading import Thread
from subprocess import check_call, CalledProcessError
from messidge import default_location
from tfnz.location import Location


def base_argparse(progname, location=True):
    parser = argparse.ArgumentParser(prog=progname)
    if location:
        connection_group = parser.add_argument_group('connection options')
        connection_group.add_argument('--location', help='use a non-default location', metavar='x.20ft.nz')
        connection_group.add_argument('--local', help='a non-dns ip for the location', metavar='x.local')
    return parser


def generic_cli(parser, implementations, *, quiet=True, location=True):
    args = parser.parse_args()
    dl = None
    loc = None

    if 'command' not in args:
        args.command = None

    if args.command is None and None not in implementations:
        parser.print_help()
        return

    # construct the location
    if location:
        try:
            dl = default_location(prefix="~/.20ft") if args.location is None else args.location
        except RuntimeError:
            print("There does not appear to be a 20ft account on this machine.", file=sys.stderr)
            sys.exit(1)

    # go
    try:
        if location:
            loc = Location(dl, location_ip=args.local, quiet=quiet)
        else:
            if not quiet:
                logging.basicConfig(level=logging.DEBUG)
        implementations[args.command](loc, args)
    except ValueError as e:
        print(str(e))
        exit(1)
    except KeyboardInterrupt:
        exit(1)
    finally:
        if loc is not None:
            loc.disconnect()
    exit(0)


# removes a flagged parameter from argv
# note: mutates argv
def remove_flagged(param, argv):
    flag_find = [a for a in enumerate(argv) if a[1] == param]
    if len(flag_find) == 1:
        del argv[flag_find[0][0]]  # flag
        del argv[flag_find[0][0]]  # and the value


def systemd(location, args, argv, preboot, cert):
    # note that this is *knee* *deep* in potential security holes but since it is connecting to
    # a non-multi-tenanted server under the user's control I'm not all that bothered about it.
    # you were warned

    # checks
    if '/' not in args.source:
        print("Use a tagged image (i.e. my/example) to create a service")
        return 1
    if '@' not in args.systemd:
        print("Please use a 'user@server' ssh connection string.")
    username = args.systemd.split('@')[0]

    # build the ssh command line
    if args.identity is None:
        ssh = ['ssh', args.systemd]
        sftp = 'sftp %s' % args.systemd
    else:
        ssh = ['ssh', '-i', args.identity, args.systemd]
        sftp = 'sftp -i %s %s' % (args.identity, args.systemd)

    # ensure file structure on the receiving machine
    service_name = args.source.replace('/', '-')
    path = '/home/%s/%s/' % (username, service_name)
    check_call(ssh + ['mkdir', '-p', path])

    # ensure the image is uploaded to the *location* (not the client-local server)
    location.ensure_image_uploaded(args.source)

    # copy any preboot files
    for source, _ in preboot:
        check_call('echo "put %s" | %s:%s' % (source, sftp, path), shell=True)

    # copy a certificate
    if cert is not None:
        check_call('echo "put %s" | %s:%s' % (cert[0], sftp, path), shell=True)
        check_call('echo "put %s" | %s:%s' % (cert[1], sftp, path), shell=True)
        if len(cert) == 3:
            check_call('echo "put %s" | %s:%s' % (cert[2], sftp, path), shell=True)

    # remove the procname, --systemd and --identity from argv
    del argv[0]
    remove_flagged('--systemd', argv)
    remove_flagged('--identity', argv)

    # render the unit file
    filename = service_name + '.service'
    with open(filename, 'w') as f:
        f.write('''
[Unit]
Description=20ft-%s

[Service]
Type=simple
Environment=PYTHONUNBUFFERED=1
ExecStart=/usr/local/bin/tf %s
WorkingDirectory=%s
KillSignal=SIGINT
TimeoutStopSec=5
Restart=always
RestartSec=5
User=%s
Group=%s

[Install]
WantedBy=multi-user.target
    ''' % (args.source, ' '.join(argv), path, username, username))
    check_call('echo "put %s" | %s:%s\n' % (filename, sftp, path), shell=True)
    check_call(['rm', filename])

    # let systemd know
    try:
        check_call(ssh + ['sudo', 'ln', '-s', path + filename, '/etc/systemd/system/'])
    except CalledProcessError as e:
        print("WARNING: creating symlink into /etc/systemd/system failed - link may well be there already")
    check_call(ssh + ['sudo', 'systemctl', 'enable', service_name])
    check_call(ssh + ['sudo', 'systemctl', 'daemon-reload'])
    check_call(ssh + ['sudo', 'systemctl', 'start', service_name])


class Interactive:
    def __init__(self, location, container):
        self.stdin_attr = None
        self.location = location
        self.container = container
        self.exit_read, self.exit_write = os.pipe()
        self.thread = Thread(target=self.stdin_loop, name="Stdin loop")
        self.thread.start()

    def stdin_loop(self):
        # fails in debugger so we'll just ignore that
        print("Interactive session - escape is triple \'^]\'.", flush=True)
        try:
            tty.setraw(sys.stdin.fileno())
        except termios.error:
            pass

        # message loop
        while True:
            ready = select.select((sys.stdin, self.exit_read), (), ())
            if self.exit_read in ready[0]:
                return
            data = sys.stdin.read(1)
            self.location.call_on_main(self.container.stdin, (data.encode(),))

    @staticmethod
    def stdout_callback(obj, out):
        # strip nasty control code things
        parts = re.split(b'\x1b\[\d*n', out, maxsplit=1)
        sys.stdout.buffer.write(parts[0] + (parts[1] if len(parts) > 1 else b''))
        sys.stdout.flush()

    def stop(self, obj=None, code=None):
        # reset stdin if we can
        os.write(self.exit_write, b'\n')



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

__all__ = ['location', 'node', 'container', 'endpoint', 'volume', 'process', 'tunnel', 'docker']

import logging
import re
import weakref
import shortuuid
from messidge.client.connection import Connection
from typing import Optional, Union, List
from base64 import b64encode
from _thread import allocate_lock


class Waitable:
    """An object that can be waited on (until marked as ready)"""
    def __init__(self, locked: Optional[bool]=True):
        self.wait_lock = allocate_lock()
        if locked:
            self.wait_lock.acquire()

    def __del__(self):
        if self.wait_lock.locked():
            self.wait_lock.release()

    def wait_until_ready(self, timeout: Optional[int]=10):
        """Blocks waiting for a (normally asynchronous) update indicating the object is ready."""
        # this lock is used for waiting on while uploading layers, needs to be long
        acquired = self.wait_lock.acquire(timeout=timeout)
        if not acquired:
            raise TimeoutError("wait_until_ready timed out")
        self.wait_lock.release()
        return self

    def mark_as_ready(self):
        if self.wait_lock.locked():
            self.wait_lock.release()

    def mark_not_ready(self):
        if not self.wait_lock.locked():
            self.wait_lock.acquire()


class Killable:
    """An object that can be marked as dead - and either bail and carry on, or raise if dead"""
    def __init__(self):
        self.dead = False

    def mark_as_dead(self):
        self.dead = True

    def bail_if_dead(self):
        # use this where the user is not really to blame
        if self.dead:
            logging.debug("Object was previously terminated (carrying on): " + self.__repr__())
        return self.dead

    def ensure_alive(self):
        # use this where the user has tried to use the object but it's kinda their fault it's no longer there
        if self.dead:
            raise ValueError("Cannot call method on a terminated object: " + self.__repr__())


class Connectable:
    """A resource that can be connected to in the private IP space"""
    def __init__(self, conn: Connection, uuid: str, node, ip):
        self.conn = weakref.ref(conn)
        self.uuid = uuid
        self.node_pk = node if isinstance(node, bytes) else node.pk
        self.ip = ip  # gets used externally, don't delete it

    def allow_connection_from(self, obj):
        self.conn().send_blocking_cmd(b'allow_connection', {'node': self.node_pk,
                                                            'container': self.uuid,
                                                            'ip': obj.ip})
        logging.info("Allowed connection (from %s) on: %s" % (obj.uuid.decode(), self.uuid))

    def disallow_connection_from(self, obj):
        self.conn().send_cmd(b'disallow_connection', {'node': self.node_pk,
                                                      'container': self.uuid,
                                                      'ip': obj.ip})
        logging.info("Disallowed connection (from %s) on: %s" % (obj.uuid.decode(), self.uuid))


class Taggable:
    """A resource that might have a tag - namespaced by user pk."""
    # user, uuid and tag are all *held* as bytes
    tag_re = re.compile(b'\A[^0-9a-z\-_.]*\Z')
    short_uuid_re = re.compile(b'\A[' + shortuuid.get_alphabet().encode() + b']{22}\Z')

    def __init__(self, user: Union[str, bytes], uuid: Union[str, bytes], tag: Optional[Union[str, bytes]]=None):
        if user is None or uuid is None:
            raise RuntimeError('Taggable resources must be constructed with at least a pk and uui')
        if isinstance(user, str):
            self.user = user.encode()
        else:
            self.user = user

        if isinstance(uuid, str):
            self.uuid = uuid.encode()
        else:
            self.uuid = uuid

        self.tag = None if tag is None else Taggable.valid_tag(tag)

    def uuid_key(self) -> (bytes, bytes):
        return self.user, self.uuid

    def tag_key(self) -> (bytes, bytes):
        """Effectively a namespaced tag."""
        if self.tag is None:
            return None
        return self.user, self.tag

    @staticmethod
    def valid_tag(tag: Union[bytes, str]) -> Union[bytes, None]:
        """Ensure that the passed tag is at least vaguely plausible - does not check for clashes"""
        if tag is None:
            return None
        if len(tag) == 0:
            raise ValueError("Tag passed for approval was blank")
        tag = tag.lower()
        if isinstance(tag, str):
            tag = tag.encode()
        if Taggable.tag_re.search(tag) is not None:
            raise ValueError("Tag names can only use 0-9 a-z - _ and .")
        if Taggable.short_uuid_re.match(tag) is not None:
            raise ValueError("Tag names cannot look like UUIDs")
        return tag

    def namespaced_display_name(self) -> str:
        return self.uuid.decode() if self.tag is None else (self.uuid.decode() + ':' + self.tag.decode())

    def global_display_name(self) -> str:
        return b64encode(self.user).decode() + ':' + self.namespaced_display_name()


class TaggedCollection:
    """A collection of taggable objects"""
    def __init__(self, initial: Optional[List[Taggable]]=None):
        self.objects = {}
        self.uuid_uuidkey = {}
        self.uniques = 0
        if initial is not None:
            for init in initial:
                self.add(init)

    def __del__(self):
        for obj in self.objects.values():
            del obj
        self.objects = {}

    # emulating a dictionary
    def __len__(self) -> int:
        return self.uniques

    def __getitem__(self, uuid: Union[str, bytes]) -> Taggable:  # applies to just UUID's
        if isinstance(uuid, str):
            uuidkey = self.uuid_uuidkey[uuid.encode()]
        else:
            uuidkey = self.uuid_uuidkey[uuid]
        return self.objects[uuidkey]

    def __contains__(self, uuid: Union[str, bytes]) -> bool:
        try:
            self[uuid]  # throws if it can't get it so, yes, this does actually do something
            return True
        except KeyError:
            return False

    def __setitem__(self, key, value):
        raise RuntimeError("Tagged collection can only be inserted to with the 'add' method")

    def __iter__(self):
        raise RuntimeError("Cannot iterate over a tagged collection")

    # what we came here for
    def add(self, obj: Taggable):
        if self.will_clash(obj.user, obj.uuid, obj.tag):
            raise RuntimeError("Cannot add to TaggedCollection because there will be a namespace clash")
        self.objects[obj.uuid_key()] = obj
        self.uuid_uuidkey[obj.uuid] = obj.uuid_key()
        if obj.tag_key() is not None:
            self.objects[obj.tag_key()] = obj
        self.uniques += 1

    def get(self, user: Union[bytes, str], key: Union[bytes, str]) -> Taggable:
        """Fetch using an ill-defined key: uuid or tag or uuid:tag"""
        if key is None:
            raise RuntimeError("Key not passed when fetching from TaggedCollection")
        if isinstance(user, str):
            user = user.encode()
        if isinstance(key, str):
            key = key.encode()

        # uuid or key or uuid:key?
        parts = key.split(b':')
        if len(parts) > 2:
            raise ValueError("Too many parts in tag: " + key.decode())

        # will match (user, uuid) or (user, tag)
        try:
            return self.objects[(user, parts[0])]
        except KeyError:
            # it seems as if we would need to search using parts[1] but...
            # we have both (user, uuid) and (user, tag) in the collection so it doesn't matter if we've passed
            # uuid, uuid:tag (matches UUID), or tag
            raise KeyError('Failed to get from a TaggedCollection with user=%s key=%s' % (user, key))

    def remove(self, obj: Taggable):
        del self.objects[obj.uuid_key()]
        del self.uuid_uuidkey[obj.uuid]
        if obj.tag_key() in self.objects:
            del self.objects[obj.tag_key()]
        self.uniques -= 1

    def will_clash(self, user: bytes, uuid: bytes, tag: bytes) -> bool:
        if tag is not None:
            if (user, tag) in self.objects:
                return True
        if (user, uuid) in self.objects:
            return True
        return False

    def values(self):
        return set(self.objects.values())  # de-dupe

    def __repr__(self):
        return "<TaggedCollection uniques=%d>" % self.uniques



# Copyright (c) 2017 David Preece, All rights reserved.
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

import os
import weakref
import logging
from paramiko import SFTPServerInterface, SFTPAttributes, SFTPHandle, SFTP_OK
from paramiko.sftp_server import SFTPServer


class Sftp(SFTPServerInterface):
    """Proxy object for a filesystem"""
    # Literally just proxies calls over to the location
    # http://docs.paramiko.org/en/2.1/api/server.html

    # These are blocking commands to give the container a chance to return an exception

    def __init__(self, transport):
        super().__init__(transport)
        self.node_pk = transport.parent().node().pk
        self.container = weakref.ref(transport.container())
        self.conn = weakref.ref(transport.container().conn())

    def list_folder(self, directory):
        ls = self.conn().send_blocking_cmd(b'ls_dir', {"node": self.node_pk,
                                                       "container": self.container().uuid,
                                                       "directory": directory})
        rtn = []
        if 'error' in ls.params:
            return SFTPServer.convert_errno(ls.params['error'])
        else:
            for entry in ls.params['entries']:
                attr = SFTPAttributes.from_stat(os.stat_result(entry[1]), entry[0])
                rtn.append(attr)
            return rtn

    def stat(self, path):
        stat = self.conn().send_blocking_cmd(b'stat_file', {"node": self.node_pk,
                                                            "container": self.container().uuid,
                                                            "filename": path})
        if 'error' in stat.params:
            return SFTPServer.convert_errno(stat.params['error'])
        else:
            return SFTPAttributes.from_stat(os.stat_result(stat.params['stat']))

    def lstat(self, path):
        stat = self.conn().send_blocking_cmd(b'lstat_file', {"node": self.node_pk,
                                                             "container": self.container().uuid,
                                                             "filename": path})
        if 'error' in stat.params:
            return SFTPServer.convert_errno(stat.params['error'])
        else:
            return SFTPAttributes.from_stat(os.stat_result(stat.params['lstat']))

    def open(self, path, flags, attr):
        return SftpFile(flags, self.node_pk, self.container, path)

    def remove(self, path):
        rm = self.conn().send_blocking_cmd(b'rm_file', {"node": self.node_pk,
                                                        "container": self.container().uuid,
                                                        "filename": path})
        if 'error' in rm.params:
            return SFTPServer.convert_errno(rm.params['error'])
        else:
            return SFTP_OK

    def rename(self, path, newpath):
        mv = self.conn().send_blocking_cmd(b'mv_file', {"node": self.node_pk,
                                                        "container": self.container().uuid,
                                                        "filename": path,
                                                        "newpath": newpath})
        if 'error' in mv.params:
            return SFTPServer.convert_errno(mv.params['error'])
        else:
            return SFTP_OK

    def mkdir(self, path, attr):
        md = self.conn().send_blocking_cmd(b'mk_dir', {"node": self.node_pk,
                                                       "container": self.container().uuid,
                                                       "directory": path})
        if 'error' in md.params:
            return SFTPServer.convert_errno(md.params['error'])
        else:
            return SFTP_OK

    def rmdir(self, path):
        rm = self.conn().send_blocking_cmd(b'rm_dir', {"node": self.node_pk,
                                                       "container": self.container().uuid,
                                                       "directory": path})
        if 'error' in rm.params:
            return SFTPServer.convert_errno(rm.params['error'])
        else:
            return SFTP_OK

    def __repr__(self):
        return "<tfnz.sftp.Sftp object at %x (container=%s)>" % (id(self), self.container().uuid.decode())


class SftpFile(SFTPHandle):
    """Proxy object for an actual file"""
    def __init__(self, flags, node_pk, container, filename):
        super().__init__(flags)
        logging.debug("Opened a file via sftp: " + filename)
        self.node_pk = node_pk
        self.container = container
        self.filename = filename
        try:
            self.data = bytearray(container().fetch(filename))
            logging.debug("...retrieved file: " + filename)
        except ValueError:
            self.data = bytearray()
            logging.debug("...will create new file on writing: " + filename)

    def read(self, offset, length):
        mv = memoryview(self.data)
        return bytes(mv[offset:offset+length])

    def write(self, offset, data):
        self.data[offset:offset+len(data)] = data
        self.container().conn().send_blocking_cmd(b'write_file', {"node": self.node_pk,
                                                                  "container": self.container().uuid,
                                                                  "filename": self.filename,
                                                                  "offset": offset}, bulk=data)
        return SFTP_OK

    def close(self):
        logging.debug("Closed a file via sftp: " + self.filename)
        return SFTP_OK

